/* vi: set sw=4 ts=4: */
/*
 * checksum(crc16) implementation for busybox
 *
 *
 * Copyright (C) [2004] by [frank chang] <frank_chang@asus.com.tw>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 */


/* frank,040116,checksum  */
/* caculate CRC-16 of files */

#include <stdio.h>
#include <stdlib.h>

#include "busybox.h"

/* data for crc 16 */
//---------------------------------------------------------------------------
int gen_tab[2][8] = {
        { 0xC0C1, 0xC181, 0xC301, 0xC601, 0xCC01, 0xD801, 0xF001, 0xA001 },
        { 0x1189, 0x2312, 0x4624, 0x8C48, 0x1081, 0x2102, 0x4204, 0x8408 }  };
int crc_tab[256];

void gen16(unsigned p)
{
    int     i, j, val;

    for (i = 0; i < 256; i++) {
        val = 0;
        for (j = 0; j < 8; j++)
            if (i & (1<<j))
                val ^= gen_tab[p][j];
        crc_tab[i] = val;
    }
}
unsigned
cal_crc16(unsigned char *data,unsigned count,unsigned old_crc)
{
    union {
        unsigned w;
        struct {
            unsigned char lo;
            unsigned char hi;
        } b;
    } crc;

    crc.w = old_crc;
    while (count-- != 0)
        crc.w = crc_tab[*data++ ^ crc.b.lo] ^ crc.b.hi;
    return crc.w;
}

int caculate_crc16(char *filename,unsigned long offset)
{
	unsigned int crc=0;
	unsigned char buffer[256];
	int size;
	FILE *fptr;
	
	fptr=fopen(filename,"rb");
	if(!fptr)
		{
		printf("open file %s failed !",filename);
		return -1;	
		}
	fseek(fptr,offset,SEEK_SET);
	while(1)
		{
		size=fread(buffer,1,256,fptr);
    	if(size > 0)
        	crc=cal_crc16(buffer, size , crc);
    	if(size < 256) break;	
		}
	fclose(fptr);
	return crc;
}
extern int checksum_main(int argc, char **argv)
{
	int crc;
	int i;
	
	if(argc < 1) show_usage();
	gen16(1);
	if(strcmp(argv[1],"all") == 0)
		{
		crc=caculate_crc16("/dev/mtdblock/2",0);		
		if(crc >= 0)
			printf("%s checksum [%04X]\n","bootrom",crc);
		crc=caculate_crc16("/dev/mtdblock/1",0);		
		if(crc >= 0)
			printf("%s checksum [%04X]\n","kernel",crc);
		crc=caculate_crc16("/dev/mtdblock/0",0);		
		if(crc >= 0)
			printf("%s checksum [%04X]\n","nsp",crc);
		/* configuration file , skip 0x2800 */	
		crc=caculate_crc16("/dev/mtdblock/3",0x2800);		
		if(crc >= 0)
			printf("%s checksum [%04X]\n","configuration",crc);
		}
	else
		{
		for(i=1;i<argc;i++)
			{	
			crc=caculate_crc16(argv[1],0);		
			if(crc >= 0)
				printf("%s checksum [%04X]\n",argv[1],crc);
			else
				break;	
			}	
		}	
}