/***************************************************************

        bwbasic.c       Main Program File
                        for Bywater BASIC Interpreter

    "I was no programmer, neither was I a
    programmer's son; but I was an herdman
    and a gatherer of sycomore fruit."
        - Amos 7:14b AV, slightly adapted

    bwBASIC version 3.03 beta
    Copyright (c) 1992-2000, Ted A. Campbell
    Copyright (c) 1995-2000, Jon B. Volkoff
    Copyright (c) 2008, TheChief (www.RouterTech.Org)
      TheChief - added
      * autoexec.bas processing
      * command line parameter processinng (argc$, and argv$)
      * resolving variables to values before calling external program

    http://www.bwBASIC.com
    email: bwBASIC@bwBASIC.com

    Copyright and Permissions Information:

    All U.S. and international rights are claimed by the authors.

    This software is released under the terms of the GNU General
    Public License (GPL), which is distributed with this software
    in the file "COPYING".  The GPL specifies the terms under
    which users may copy and use the software in this distribution.
    Please note that it rules out the distribution of proprietary
    versions of the software, and it rules out the use of altered
    versions of the software. This means that you cannot remove
    the Bywater BASIC name or copyright information.

    A separate license is available for proprietary or commercial
    distribution, for information on which you should contact
    bwBASIC@bwBASIC.com.

***************************************************************/

/*---------------------------------------------------------------*/
/* NOTE: Modifications marked "JBV" were made by Jon B. Volkoff, */
/* 11/1995 (eidetics@cerf.net).                                  */
/*---------------------------------------------------------------*/

/*-----------------------------------------------------------*/
/* Modifications marked "MY" were made by Mark Young, 4/1998 */
/* (MarkYoung@tx.slr.com).                                   */
/*-----------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <math.h>

#include "bwbasic.h"
#include "bwb_mes.h"

#include <signal.h>

#include <setjmp.h>

char *bwb_ebuf;				/* error buffer */
static char *read_line;
int bwb_trace = FALSE;

/* Was = stderr (JBV 8/98) */
FILE *errfdevice;                       /* output device for error messages */

#if HAVE_LONGJUMP
jmp_buf mark;
#endif

static int program_run = 0;		/* has the command-line program been run? */
int bwb_curtask = 0;			/* current task */

struct bwb_variable *pr;                /* BWB.PROMPT$ variable */
struct bwb_variable *im;                /* BWB.IMPLEMENTATION$ variable */
struct bwb_variable *co;                /* BWB.COLORS variable */
struct bwb_variable *cm;                /* COMMAND$ variable (JBV 10/18/97) */

// thechief
struct bwb_variable *paramc;
struct bwb_variable *params [MAX_PARMS];
char *paramnames [MAX_PARMS];
// end: thechief

char progfile[ MAXARGSIZE ];		/* program file */
int rescan = TRUE;                      /* program needs to be rescanned */
int number = 0;				/* current line number */
struct bwb_line *bwb_l;			/* current line pointer */
struct bwb_line bwb_start;		/* starting line marker */
struct bwb_line bwb_end;		/* ending line marker */
struct bwb_line *data_line;     	/* current line to read data */
int data_pos = 0;                       /* position in data_line */
struct bwb_variable var_start;		/* variable list start marker */
struct bwb_variable var_end;		/* variable list end marker */
struct bwb_function fnc_start;		/* function list start marker */
struct bwb_function fnc_end;		/* function list end marker */
struct fslte fslt_start;		/* function-sub-label lookup table start marker */
struct fslte fslt_end;			/* function-sub-label lookup table end marker */
int exsc = -1;				/* EXEC stack counter */
int expsc = 0;				/* expression stack counter */
int xtxtsc = 0;				/* eXecute TeXT stack counter */
struct exse *excs;			/* EXEC stack */
struct exp_ese *exps;			/* Expression stack */
struct xtxtsl *xtxts;                   /* Execute Text stack */

/*----------------------------------------------*/
/* Added by JBV 10/20/97 to support CONT/RESUME */
/*----------------------------------------------*/
int old_exsc;
struct exse old_excs[EXECLEVELS];
int break_detected = 0;
int cont_resume_pending = 0;

/* Prototypes for functions visible only to this file */

extern int is_ln( char *buffer );

// **** thechief: provide access to command line arguments, other than from COMMAND$ **//
/*
char* itoa(int val, int base)
{
	static char buf[32] = {0};
	int i = 30;	
	for(; val && i ; --i, val /= base)
		buf[i] = "0123456789abcdef"[val % base];	
	return &buf[i+1];	
}
*/

void strreverse(char* begin, char* end) 
{	
	char aux;	
	while(end>begin)
		aux=*end, *end--=*begin, *begin++=aux;
}
	
char *itoa (int value, char* str, int base) 
{
	static char num[] = "0123456789abcdefghijklmnopqrstuvwxyz";
	char* wstr=str;
	int sign;
	div_t res;
	
	// Validate base	
	if (base<2 || base>35){ *wstr='\0'; return NULL; }	
	
	// Take care of sign
	if ((sign=value) < 0) value = -value;
	
	// Conversion. Number is reversed.
	do 
	{	
	  res = div(value,base);
	  *wstr++ = num[res.rem];
	}
	while (value=res.quot);
	
	if (sign < 0) *wstr++ ='-';
	*wstr='\0';
		
	// Reverse string
	strreverse(str,wstr-1);
	return str;
}
	
int doparamstrs( int argc, char **argv )
{
 int i;  
 int j;
 int x;
 char b[16];
 char buf[8];
  
   // sort out the variable names 
   for (i = 0; i < MAX_PARMS; i++)
   {
     strcpy (b, DEFVNAME_PARAM_S);
     strcat (b, itoa (i, &buf, 10));
     strcat (b, "$");
     paramnames[i] = strdup (b);
   }

//   for (i = 0; i < MAX_PARMS; i++) { printf("%i=%s\n",i,paramnames[i]);}  
   
   x = argc;
   x -= 2;
   if (x >= MAX_PARMS-1) { x = MAX_PARMS-2; }
   
   // process the number of parameters given
   paramc = var_find ( DEFVNAME_PARAMC );
   paramc->preset = TRUE;
   paramc->common = TRUE;
   str_ctob ( var_findsval ( paramc, paramc->array_pos ),  itoa (x, &buf, 10));
   
   // process the parameter strings   
   // first set the names
   for (i=0; i < MAX_PARMS; i++)
   {
     params [i] = var_find ( paramnames [i] );
     params [i]->preset = TRUE;
     params [i]->common = TRUE;
   }

   if (argc < 2) return 1;
   
   // the name of the script itself (param_str0)
   str_ctob ( var_findsval ( params [0], params [0]->array_pos ), argv [ 1 ] );
    
   // then set the values
   if ( x > 0 ) 
   { 
     for (i = 1; i <= x; i++)
     {
       j = i + 1;
//       printf("params %i = %s, and j=%i\n", i, argv[j], j);
       if ((j < argc) && (j < MAX_PARMS)) 
       {
         str_ctob ( var_findsval ( params [i], params [i]->array_pos ), argv [ j ] );
       } 
       else break;
     }
   }
   /*
   if ( argc > 2 ) 
   { 
     for (i = 0; i < argc; i++)
     {
       j = i + 1;
       if ((j < argc) && (j < MAX_PARMS)) 
       {
         str_ctob ( var_findsval ( params [i], params [i]->array_pos ), argv [ j ] );
       } 
       else break;
     }
   }
   */
   return 0;
}
//* end: thechief

/***************************************************************

        FUNCTION:       bwb_init()

        DESCRIPTION:    This function initializes bwBASIC.

***************************************************************/

void
bwb_init( int argc, char **argv )
   {
   static FILE *input = NULL;
   register int n;
   static int ran_profile = FALSE;
#if PROFILE
   struct bwb_variable *v;
#endif

#if RT_AUTOEXEC /* thechief */
   struct bwb_variable *vrt;
#endif /* thechief */

#if REDIRECT_STDERR
   FILE *newerr;
#endif

#if PROFILE
   FILE *profile;
   static int profile_loaded = FALSE; /* MY: added */
#endif

   static char start_buf[] = "\0";
   static char end_buf[] = "\0";

#if INTENSIVE_DEBUG
   prn_xprintf( stderr, "Memory Allocation Statistics:\n" );
   prn_xprintf( stderr, "----------------------------\n" );
#endif

   /* set all task pointers to NULL */

   /* Memory allocation */
   /* eXecute TeXT stack */

   /* Revised to CALLOC0 pass-thru call by JBV */
   if ( ( xtxts = (struct xtxtsl *) CALLOC0( XTXTSTACKSIZE, sizeof( struct xtxtsl ), "bwb_init") ) == NULL )
      {
#if PROG_ERRORS
      bwb_error( "in bwb_init(): failed to find memory for xtxts" );
#else
      bwb_error( err_getmem );
#endif
      }

   /* expression stack */

   /* Revised to CALLOC0 pass-thru call by JBV */
   if ( ( exps = (struct exp_ese *) CALLOC0( ESTACKSIZE, sizeof( struct exp_ese ), "bwb_init") ) == NULL )
      {
#if PROG_ERRORS
      bwb_error( "in bwb_init(): failed to find memory for exps" );
#else
      bwb_error( err_getmem );
#endif
      }

   /* EXEC stack */

   /* Revised to CALLOC0 pass-thru call by JBV */
   if ( ( excs = (struct exse *) CALLOC0( EXECLEVELS, sizeof( struct exse ), "bwb_init") ) == NULL )
      {
#if PROG_ERRORS
      bwb_error( "in bwb_init(): failed to find memory for excs" );
#else
      bwb_error( err_getmem );
#endif
      }

   /* initialize tables of variables, functions */

   bwb_start.number = 0;
   bwb_start.next = &bwb_end;
   bwb_end.number = MAXLINENO + 1;
   bwb_end.next = &bwb_end;
   bwb_start.buffer = start_buf;
   bwb_end.buffer = end_buf;
   data_line = &bwb_start;
   data_pos = 0;
   exsc = -1;
   expsc = 0;
   xtxtsc = 0;
   bwb_start.position = 0;
   bwb_l = &bwb_start;

   var_init( 0 );
   fnc_init( 0 );
   fslt_init( 0 );

   /* character buffers */

   /* Revised to CALLOC0 pass-thru call by JBV */
   if ( ( bwb_ebuf = (char *) CALLOC0( MAXSTRINGSIZE + 1, sizeof(char), "bwb_init") ) == NULL )
      {
#if PROG_ERRORS
      bwb_error( "in bwb_init(): failed to find memory for bwb_ebuf" );
#else
      bwb_error( err_getmem );
#endif
      }
   /* Revised to CALLOC0 pass-thru call by JBV */
   if ( ( read_line = (char *) CALLOC0( MAXREADLINESIZE + 1, sizeof(char), "bwb_init") ) == NULL )
      {
#if PROG_ERRORS
      bwb_error( "in bwb_init(): failed to find memory for read_line" );
#else
      bwb_error( err_getmem );
#endif
      }

#if TEST_BSTRING
   for ( n = 0; n < ESTACKSIZE; ++n )
      {
      sprintf( exps[ n ].sval.name, "<Exp stack bstring %d>", n );
      }
#endif

   /* assign memory for the device table */

   /* Revised to CALLOC0 pass-thru call by JBV */
   if ( ( dev_table = (struct dev_element *) CALLOC0( DEF_DEVICES, sizeof( struct dev_element ), "bwb_init") ) == NULL )
      {
#if PROG_ERRORS
      bwb_error( "in bwb_init(): failed to find memory for dev_table" );
#else
      bwb_error( err_getmem );
#endif
      bwx_terminate();
      }

   /* initialize all devices to DEVMODE_AVAILABLE */

   for ( n = 0; n < DEF_DEVICES; ++n )
      {
      dev_table[ n ].mode = DEVMODE_AVAILABLE;
      dev_table[ n ].reclen = -1;
      dev_table[ n ].cfp = NULL;
      dev_table[ n ].buffer = NULL;
      dev_table[ n ].width = DEF_WIDTH;
      dev_table[ n ].col = 1;
      }

   pr = var_find( DEFVNAME_PROMPT );
   pr->preset = TRUE;
   pr->common = TRUE;
   str_ctob( var_findsval( pr, pr->array_pos ), PROMPT );

   /*-----------------------*/
   /* Added by JBV 10/18/97 */
   /*-----------------------*/
   cm = var_find( DEFVNAME_CMDARG );
   cm->preset = TRUE;
   cm->common = TRUE;
   if ( argc > 2 ) str_ctob( var_findsval( cm, cm->array_pos ), argv[ 2 ] );

   // thechief
   doparamstrs (argc, argv);
   /* Signon message */

   if (argc == 1) bwx_signon(); /* Added argc check (JBV 8/98) */

   /* Redirect stderr if specified */

#if REDIRECT_STDERR
   newerr = freopen( ERRFILE, "w", stderr );
   if ( newerr == NULL )
      {
      sprintf( bwb_ebuf, "Failed to redirect error messages to file <%s>\n",
         ERRFILE );
      errfdevice = stdout;
      prn_xprintf( errfdevice, bwb_ebuf );
      }
   else
      {
      sprintf( bwb_ebuf, "NOTE: Error messages are redirected to file <%s>\n",
         ERRFILE );
      prn_xprintf( errfdevice, bwb_ebuf );
      errfdevice = stderr;
      }
#else
   /* Normally, error output is to stdout; this allows us to 
      count columns, etc. */ 

   errfdevice = stdout;
#endif

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in main(): Ready to save jump MARKER" );
   bwb_debug( bwb_ebuf );
#endif

   /* set a buffer for jump: program execution returns to this point
      in case of a jump (error, interrupt, or finish program) */

#if INTERACTIVE

   /* Interrupt handler was break_mes (JBV 7/98) */
   signal( SIGINT, break_grabber ); /* MY: signal() ignored in prot mode - see raise() calls */

   setjmp( mark );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_init(): Return from jump MARKER, program run <%d>",
      program_run + 1 );
   bwb_debug( bwb_ebuf );
   getchar();
#endif

#endif		/* INTERACTIVE */

      
/******** thechief *********************/
/* 0: run autoexec.bas */
#if RT_AUTOEXEC
      if ( ( profile = fopen( AUTOEXECNAME, "r" )) != NULL )
         {
         bwb_fload( profile );	/* load file */
         bwb_run( &bwb_start );	/* run file */
#if THEOLDWAY
         while ( exsc > -1 )
              {
              bwb_execline();
              }
#else
         bwb_mainloop();
#endif
         /* mark all profiled variables as preset */
         for ( vrt = var_start.next; vrt != &var_end; vrt = vrt->next )
            {
#if INTENSIVE_DEBUG
            sprintf( bwb_ebuf, "in bwb_init(): marked variable <%s> preset TRUE",
               vrt->name );
            bwb_debug( bwb_ebuf );
#endif
            vrt->preset = TRUE;
            }
         bwb_new( &bwb_start ); 	/* remove profile from memory */
         }
#endif
/******** end: thechief *********************/

/*    There are three basic places to run a program file: a profile.bas, 
      a program specified as an argument, or a program entered from the 
      interactive command line. We now proceed to consider these three
      possibilities. */ 
 
/* 1. If there is a profile.bas and no argument, execute it */

#if PROFILE
   /* MY: added code to exec profile only once AND moved the
    * profile loading AFTER setjmp so if profile has an error and
    * calls bwb_error and then longjmp, then there will be a valid
    * mark to jump TO!
    */

   if (( profile_loaded == FALSE ) && ( argc < 2 )) /* MY: added */
      { /* MY: added */
      profile_loaded = TRUE; /* MY: added */
      if ( ( profile = fopen( PROFILENAME, "r" )) != NULL )
         {
         bwb_fload( profile );	/* load profile */
         bwb_run( &bwb_start );	/* run profile */

         /* profile must be run immediately, not by scheduler */
         /* AGENDA: all program loops need to be regularized */

#if THEOLDWAY
         while ( exsc > -1 )
              {
              bwb_execline();
              }
#else
         bwb_mainloop();
#endif

         /* mark all profiled variables as preset */

         for ( v = var_start.next; v != &var_end; v = v->next )
            {

#if INTENSIVE_DEBUG
            sprintf( bwb_ebuf, "in bwb_init(): marked variable <%s> preset TRUE",
               v->name );
            bwb_debug( bwb_ebuf );
#endif

            v->preset = TRUE;
            }

         bwb_new( &bwb_start ); 	/* remove profile from memory */
         ran_profile = TRUE; 
         }
      } /* MY: added */
#endif


#if INTERACTIVE
#else

   /* 2. If INTERACTIVE is FALSE and we did not run a profile, 
	then we must have a program file */

   if (( argc < 2 ) && (ran_profile == FALSE))
      {
      bwb_error( err_noprogfile );
      }

   /* if INTERACTIVE is FALSE and we did run a profile, 
	then we must exit */

   if (( argc < 2 ) && (ran_profile == TRUE))
      {
      bwx_terminate();
      }

#endif				/* INTERACTIVE */

   /* 2. This is option 2 again -- run a program specified as an argument,
      but here it appears as part of the INTERACTIVE environment */ 

   /* check to see if there is a program file: but do this only the first
      time around! */

   ++program_run;
   if (( argc > 1 ) && ( program_run == 1 ))
      {
      strcpy( progfile, argv[ 1 ] ); /* JBV */
      if ( ( input = fopen( progfile, "r" )) == NULL ) /* JBV */
         {
         strcat( progfile, ".bas" );
         if ( ( input = fopen( progfile, "r" )) == NULL )
            {
            progfile[ 0 ] = 0;
            sprintf( bwb_ebuf, err_openfile, argv[ 1 ] );
            bwb_error( bwb_ebuf );
            }
         }
      if ( input != NULL )
         {
         /* strcpy( progfile, argv[ 1 ] ); */  /* Removed by JBV */
#if INTENSIVE_DEBUG
         sprintf( bwb_ebuf, "in main(): progfile is <%s>.", progfile );
         bwb_debug( bwb_ebuf );
#endif

         bwb_fload( input );
         bwb_run( &bwb_start );

         /* if not interactive, then terminate after running this program */

#if INTERACTIVE
#else
         /* file to be run immediately, not by scheduler */
         /* AGENDA: Here's another program loop */ 

#if THEOLDWAY
         while ( exsc > -1 )
              {
              bwb_execline();
              }
#else
         bwb_mainloop();
#endif

         bwx_terminate();
#endif

         }
      }

   }

/***************************************************************

        FUNCTION:       bwb_interact()

        DESCRIPTION: 	This function gets a line from the user
			and processes it.

***************************************************************/

#if INTERACTIVE
int
bwb_interact( void )
   {
   char tbuf[ MAXSTRINGSIZE + 1 ]; /* JBV */

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_interact(): ready to read from keyboard" );
   bwb_debug( bwb_ebuf );
#endif

   /* take input from keyboard */

   read_line[0] = '\0'; /* JBV 10/20/97 */
   bwb_gets( read_line );
   bwb_stripcr( read_line ); /* JBV */

   /* If there is no line number, execute the line as received */

   if ( is_ln( read_line ) == FALSE )
      {
      bwb_xtxtline( read_line );
      }

   /*-----------------------------------------------------------------*/
   /* Another possibility: if read_line is a numeric constant, delete */
   /* the indicated line number (JBV)                                 */
   /*-----------------------------------------------------------------*/
   else if ( is_numconst( read_line ) == TRUE )
   {
       strcpy(tbuf, read_line);
       sprintf(read_line, "delete %s\0", tbuf);
       bwb_xtxtline( read_line );
   }

   /* If there is a line number, add the line to the file in memory */

   else
      {
      bwb_ladd( read_line, TRUE );
#if INTENSIVE_DEBUG
      bwb_debug( "Return from bwb_ladd()" );
#endif
      }

   return TRUE;

   }

#endif				/* INTERACTIVE == TRUE */

/***************************************************************

        FUNCTION:       bwb_fload()

        DESCRIPTION: 	This function loads a BASIC program
			file into memory given a FILE pointer.

***************************************************************/

int
bwb_fload( FILE *file )
   {

   while ( feof( file ) == FALSE )
      {
      read_line[ 0 ] = '\0';
      fgets( read_line, MAXREADLINESIZE, file );
      if ( file == stdin )
         {
         * prn_getcol( stdout ) = 1;		/* reset column */
         }
      bwb_stripcr( read_line );

      /* be sure that this is not EOF with a NULL line */

      if (( feof( file ) == FALSE ) || ( strlen( read_line ) > 0 ))
	 {
	 bwb_ladd( read_line, FALSE );
	 }
      }

   /* close file stream */

   fclose( file );

   /* try to call SUB MAIN */ 

   bwb_callmain(); 

   /* return */ 

   return TRUE;
   }

/***************************************************************

        FUNCTION:       bwb_ladd()

        DESCRIPTION:    This function adds a new line (in the
                        buffer) to the program in memory.

***************************************************************/

int
bwb_ladd( char *buffer, int replace )
   {
   struct bwb_line *l, *previous, *p;
   static char *s_buffer;
   static int init = FALSE;
   static int prev_num = 0;
   char *newbuffer;

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_ladd(): add line <%s>",
      buffer );
   bwb_debug( bwb_ebuf );
#endif

   /* get memory for temporary buffer if necessary */

   if ( init == FALSE )
      {
      init = TRUE;

      /* Revised to CALLOC0 pass-thru call by JBV */
      if ( ( s_buffer = (char *) CALLOC0( (size_t) MAXSTRINGSIZE + 1, sizeof( char ), "bwb_ladd" )) == NULL )
	 {
#if PROG_ERRORS
	 bwb_error( "in bwb_ladd(): failed to find memory for s_buffer" );
#else
	 bwb_error( err_getmem );
#endif
	 return FALSE;
         }
      }

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_ladd(): s_buffer initialized " );
   bwb_debug( bwb_ebuf );
#endif

   /* get memory for this line */

   /* Revised to CALLOC0 pass-thru call by JBV */
   if ( ( l = (struct bwb_line *) CALLOC0( (size_t) 1, sizeof( struct bwb_line ), "bwb_ladd")) == NULL )
      {
#if PROG_ERRORS
      bwb_error( "in bwb_ladd(): failed to find memory for new line" );
#else
      bwb_error( err_getmem );
#endif
      return FALSE;
      }

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_ladd(): got memory." );
   bwb_debug( bwb_ebuf );
#endif

   /* note that line is not yet marked and the program must be rescanned */

   l->marked = FALSE;
   rescan = TRUE;		/* program needs to be scanned again */
   l->xnum = FALSE;

   /* get the first element and test for a line number */

   adv_element( buffer, &( l->position ), s_buffer );

   /* set line number in line structure */

   if ( is_numconst( s_buffer ) == TRUE )
      {

      l->number = atoi( s_buffer );

#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "in bwb_ladd(): line is numbered, number is <%d>",
	 l->number );
      bwb_debug( bwb_ebuf );
#endif

      prev_num = l->number;
      l->xnum = TRUE;
      ++( l->position );
      newbuffer = &( buffer[ l->position ] );

      /* allocate memory and assign buffer to line buffer */

      ln_asbuf( l, newbuffer );

      }

   /* There is not a line number */

   else
      {

#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "in bwb_ladd(): line is not numbered, using prev <%d>",
         prev_num );
      bwb_debug( bwb_ebuf );
#endif

      newbuffer = buffer;

      /* allocate memory and assign buffer to line buffer */

      ln_asbuf( l, newbuffer );

      l->xnum = FALSE;
      l->number = prev_num;
      }

   /* find the place of the current line */

   for ( previous = &bwb_start; previous != &bwb_end; previous = previous->next )
      {

      /* replace a previously existing line */

      if ( ( l->xnum == (char) TRUE ) /* Better recast this one (JBV) */
         && ( previous->number == l->number )
	 && ( replace == TRUE )
	 )
         {

#if INTENSIVE_DEBUG
         sprintf( bwb_ebuf, "in bwb_ladd(): writing to previous number <%d>",
            l->number );
         bwb_debug( bwb_ebuf );
#endif

         /* allocate memory and assign buffer to line buffer */

	 ln_asbuf( previous, newbuffer );

         /* free the current line */

         /* Revised to FREE0 pass-thru calls by JBV */
         /* if (l->buffer != NULL) FREE0( l->buffer, "bwb_ladd" ); */
         /* FREE0( l, "bwb_ladd" ); */
         bwb_freeline( l ); /* JBV */

         /* and return */

         return TRUE;

         }

      /* add after previously existing line: this is to allow unnumbered
         lines that follow in sequence after a previously numbered line */

      else if (( previous->number == l->number )
	 && ( replace == FALSE )
	 )
         {
#if INTENSIVE_DEBUG
	 sprintf( bwb_ebuf, "in bwb_ladd(): adding doubled number <%d>",
	    l->number );
	 bwb_debug( bwb_ebuf);
#endif

         /* if there are multiple instances of this particular line number,
            then it is incumbent upon us to find the very last one */

         for ( p = previous; p->number == l->number; p = p->next )
            {
#if INTENSIVE_DEBUG
	    bwb_debug( "in bwb_ladd(): advancing..." );
#endif
            previous = p;
            }

         l->next = previous->next;
         previous->next = l;
         return TRUE;
         }

      /* add a new line */

      else if ( ( previous->number < l->number )
         && ( previous->next->number > l->number ))
         {
         l->next = previous->next;
         previous->next = l;

#if INTENSIVE_DEBUG
	 sprintf( bwb_ebuf, "in bwb_ladd(): added new line <%d> buffer <%s>",
	    l->number, l->buffer );
	 bwb_debug( bwb_ebuf );
#endif

         return TRUE;
         }

      }

   /* attempt to link line number has failed; free memory */

   /* Revised to FREE0 pass-thru calls by JBV */
   /* if (l->buffer != NULL) FREE0( l->buffer, "bwb_ladd" ); */
   /* FREE0( l, "bwb_ladd" ); */
   bwb_freeline( l ); /* JBV */

   sprintf( bwb_ebuf, ERR_LINENO );
   bwb_error( bwb_ebuf );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_ladd(): attempt to add line has failed" );
   bwb_debug( bwb_ebuf );
#endif

   return FALSE;

   }

/***************************************************************

        FUNCTION:       bwb_xtxtline()

        DESCRIPTION:    This function executes a text line, i.e.,
                        places it in memory and then relinquishes
                        control.

***************************************************************/

struct bwb_line *
bwb_xtxtline( char *buffer )
   {
   struct bwb_line *c;
   char *p;
   int loop;

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_xtxtline(): received <%s>", buffer );
   bwb_debug( bwb_ebuf );
#endif

   /* increment xtxt stack counter */

   if ( xtxtsc >= XTXTSTACKSIZE )
      {
      sprintf( bwb_ebuf, "Exceeded maximum xtxt stack <%d>",
         xtxtsc );
      return &bwb_end;
      }

   ++xtxtsc;

   /* advance past whitespace */

   p = buffer;
   loop = TRUE;
   while( loop == TRUE )
      {

      switch( *p )
         {
         case '\0':                     /* end of string */

#if INTENSIVE_DEBUG
            sprintf( bwb_ebuf, "Null command line received." );
            bwb_debug( bwb_ebuf );
#endif
            --xtxtsc;
            return &bwb_end;
         case ' ':                      /* whitespace */
         case '\t':
            ++p;
            break;
         default:
            loop = FALSE;
            break;
         }

      }

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_xtxtline(): ready to get memory" );
   bwb_debug( bwb_ebuf );
#endif

   /* copy the whole line to the line structure buffer */

   ln_asbuf( &( xtxts[ xtxtsc ].l ), buffer );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_xtxtline(): copied to line buffer <%s>.",
      xtxts[ xtxtsc ].l.buffer );
   bwb_debug( bwb_ebuf );
#endif

   /* set line number in line structure */

   xtxts[ xtxtsc ].l.number = 0;
   xtxts[ xtxtsc ].l.marked = FALSE;

   /* execute the line as BASIC command line */

   xtxts[ xtxtsc ].l.next = &bwb_end;
   c = &( xtxts[ xtxtsc ].l );
   c->position = 0;

#if THEOLDWAY
   do
      {
      c = bwb_xline( c );
      }

   while( c != &bwb_end );
#endif

   bwb_incexec();		/* increment EXEC stack */
   bwb_setexec( c, 0, EXEC_NORM );		/* and set current line in it */

   /* decrement xtxt stack counter ??? */

   --xtxtsc;

   return c;

   }

/***************************************************************

        FUNCTION:       bwb_incexec()

        DESCRIPTION:    This function increments the EXEC
			stack counter.

***************************************************************/

extern void
bwb_incexec( void )
   {
   ++exsc;

   if ( exsc >= EXECLEVELS )
      {
      --exsc;
#if PROG_ERRORS
      sprintf( bwb_ebuf, "in bwb_incexec(): incremented EXEC stack past max <%d>",
         EXECLEVELS );
      bwb_error( bwb_ebuf );
#else
      bwb_error( err_overflow );
#endif
      }

   excs[ exsc ].while_line = NULL;
   excs[ exsc ].wend_line  = NULL;
   excs[ exsc ].n_cvs = 0;
   excs[ exsc ].local_variable = NULL;

   }

/***************************************************************

        FUNCTION:       bwb_decexec()

        DESCRIPTION:    This function decrements the EXEC
			stack counter.

***************************************************************/

extern void
bwb_decexec( void )
   {

   /* decrement the exec stack counter */

   --exsc;

   if ( exsc <= -1 ) /* MY 7/98: was < */
      {
      exsc = -1;
#if PROG_ERRORS
      sprintf( bwb_ebuf, "in bwb_decexec(): decremented EXEC stack past min <-1>" );
      bwb_error( bwb_ebuf );
#else
      bwb_error( err_overflow );
#endif
      }

   /* check for EXEC_ON and decrement recursively */

   if ( excs[ exsc ].code == EXEC_ON )
      {

      /* Revised to FREE0 pass-thru calls by JBV */
      /* FREE0( excs[ exsc ].while_line->buffer, "bwb_decexec" ); */
      /* FREE0( excs[ exsc ].while_line, "bwb_decexec" ); */
      bwb_freeline( excs[ exsc ].while_line ); /* JBV */

      bwb_decexec();
      }

   }

/***************************************************************

        FUNCTION:       bwb_setexec()

        DESCRIPTION:    This function sets the line and position
			for the next call to bwb_execline();

***************************************************************/

extern int
bwb_setexec( struct bwb_line *l, int position, int code )
   {

   excs[ exsc ].line = l;
   excs[ exsc ].position = position;
   excs[ exsc ].code = code;

   return TRUE;

   }

/***************************************************************

        FUNCTION:       bwb_iteration()

        DESCRIPTION:    This C function performs one iteration
                        of the interpreter. In a non-preemptive
                        scheduler, this function should be called
                        by the scheduler, not by bwBASIC code.

***************************************************************/

void
bwb_iteration( void )
   {
   if ( exsc > -1 )
      {
      /*------------------------------------------------------------*/
      /* Following section added by JBV 7/98 to support CONT/RESUME */
      /*------------------------------------------------------------*/
      if (break_detected)
         {
         break_mes();
         break_detected = 0;

#if INTENSIVE_DEBUG
         sprintf( bwb_ebuf, "in bwb_iteration(): Saving old stacks" );
         bwb_debug( bwb_ebuf );
#endif

         cont_resume_pending = 1;
         old_exsc = exsc;
         memcpy(old_excs, excs, sizeof(struct exse) * EXECLEVELS);

         /* reset all stack counters */

         exsc = -1;
         expsc = 0;
         xtxtsc = 0;

         return;
         }

      bwb_execline();			/* execute one line of program */
      }
#if INTERACTIVE
   else
      {
      bwb_interact();			/* get user interaction */
      if (break_detected)
         {
         break_mes();
         break_detected = 0; /* To support CONT/RESUME (JBV 7/98) */
         }
      }
#endif
   }

/***************************************************************

        FUNCTION:       bwb_execline()

        DESCRIPTION:    This function executes a single line of
                        a program in memory. This function is
			called by bwb_iteration().

***************************************************************/

void
bwb_execline( void )
   {
   struct bwb_line *r, *l;
   char tbuf[ MAXSTRINGSIZE + 1 ];
   
   l = excs[ exsc ].line;

   /* if the line is &bwb_end, then break out of EXEC loops */

   if ( l == &bwb_end )
      {
      exsc = -1;
      return;
      }

   /* Check for wacko line numbers  */

#if INTENSIVE_DEBUG
   if ( l->number < -1 )
      {
#if PROG_ERRORS
      sprintf( bwb_ebuf, "in bwb_execline(): received line number <%d> < -1",
         l->number );
      bwb_error( bwb_ebuf );
#else
      bwb_error( err_syntax );
#endif
      return;
      }

   if ( l->number > MAXLINENO )
      {
#if PROG_ERRORS
      sprintf( bwb_ebuf, "in bwb_execline(): received line number <%d> > MAX <%d>",
         l->number, MAXLINENO );
      bwb_error( bwb_ebuf );
#else
      bwb_error( err_syntax );
#endif
      return;
      }
#endif

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_execline(): buffer <%s>",
      &( l->buffer[ l->position ] ) );
   bwb_debug( bwb_ebuf );
#endif

   /* Print line number if trace is on */

   if ( bwb_trace == TRUE )
      {
#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "[ %d ]", l->number );
      prn_xprintf( errfdevice, bwb_ebuf );
#else
      if ( l->number > 0 )
         {
         sprintf( bwb_ebuf, "[ %d ]", l->number );
	 prn_xprintf( errfdevice, bwb_ebuf );
         }
#endif
      }

   /* Set current line for error/break handling */

   number = l->number;
   bwb_l = l;

   /* advance beyond whitespace */

   adv_ws( l->buffer, &( l->position ) );

   /*  advance past segment delimiter and warn */

#if MULTISEG_LINES
   if ( l->buffer[ l->position ] == ':' )
      {
      ++( l->position );
      adv_ws( l->buffer, &( l->position ) );
      }
   l->marked = FALSE;
#else
#if PROG_ERRORS
   if ( l->buffer[ l->position ] == ':' )
      {
      ++( l->position );
      adv_ws( l->buffer, &( l->position ) );
      sprintf( bwb_ebuf, "Enable MULTISEG_LINES for multi-segmented lines",
         VERSION );
      bwb_error( bwb_ebuf );
      }
#endif
#endif

   /* set positions in buffer */

#if MULTISEG_LINES
   if ( ( l->marked != TRUE ) || ( l->position > l->startpos ))
      {
      line_start( l->buffer, &( l->position ), &( l->lnpos ), &( l->lnum ),
         &( l->cmdpos ), &( l->cmdnum ), &( l->startpos ) );
      l->marked = TRUE;
      }
   else
      {
#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "in bwb_execline(): line <%d> is already marked",
          l->number );
      bwb_debug( bwb_ebuf );
#endif
      }
   l->position = l->startpos;
#else				/* not MULTISEG_LINES */
   line_start( l->buffer, &( l->position ), &( l->lnpos ), &( l->lnum ),
      &( l->cmdpos ), &( l->cmdnum ), &( l->startpos ) );
   if ( l->position < l->startpos )
      {
      l->position = l->startpos;
      }
#endif

   /* if there is a BASIC command in the line, execute it here */

   if ( l->cmdnum > -1 )
      {

#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "in bwb_execline(): executing <%s>", l->buffer );
      bwb_debug( bwb_ebuf );
#endif

      /* execute the command vector */

      r = bwb_cmdtable[ l->cmdnum ].vector ( l );

      }

   /* No BASIC command; try to execute it as a shell command */

#if COMMAND_SHELL
   else
      {

#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "Breaking out to shell, line num <%d> buf <%s> cmd <%d> pos <%d>",
         l->number, &( l->buffer[ l->position ] ), l->cmdnum, l->position );
      bwb_debug( bwb_ebuf );
      getchar();
#endif

      bwx_shell( l );

      /*----------------------------------------------------------*/
      /* Following bug found by Mark Young (MarkYoung@tx.slr.com) */
      /* End of line is assumed when there is a shell command, so */
      /* set the position now before we return (JBV)              */
      /*----------------------------------------------------------*/
      l->next->position = 0; /* MY: added */

      bwb_setexec( l->next, 0, excs[ exsc ].code );
      return;
      }

#else				/* COMMAND_SHELL == FALSE */

   else
     {
     bwb_error( err_uc );
     }

#endif

   /* check for end of line: if so, advance to next line and return */

   adv_ws( r->buffer, &( r->position ) );
   switch( r->buffer[ r->position ] )
      {
      case '\n':
      case '\r':
      case '\0':

#if INTENSIVE_DEBUG
	 sprintf( bwb_ebuf, "in bwb_execline(): detected end of line" );
	 bwb_debug( bwb_ebuf );
#endif

	 r->next->position = 0;
         bwb_setexec( r->next, 0, excs[ exsc ].code );
         return;			/* and return */
      }

   /* else reset with the value in r */

   bwb_setexec( r, r->position, excs[ exsc ].code );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_execline(): exit setting line number <%d>",
      r->number );
   bwb_debug( bwb_ebuf );
#endif

   }

/***************************************************************

        FUNCTION:       ln_asbuf()

        DESCRIPTION:    This function allocates memory and copies
			a null-terminated string to a line buffer.

***************************************************************/

int
ln_asbuf( struct bwb_line *l, char *s )
   {

/* Reinstated by JBV */
   if ( l->buffer != NULL )
      {
      /* Revised to FREE0 pass-thru call by JBV */
      FREE0( l->buffer, "ln_asbuf" );
      l->buffer = NULL; /* JBV */
      }

   /* Revised to CALLOC0 pass-thru call by JBV */
   if ( ( l->buffer = (char *) CALLOC0( strlen( s ) + 2, sizeof( char ), "ln_asbuf") )
      == NULL )
      {
#if PROG_ERRORS
      bwb_error( "in ln_asbuf(): failed to find memory for new line" );
#else
      bwb_error( err_getmem );
#endif
      return FALSE;
      }

   /* copy the whole line to the line structure buffer */

   strcpy( l->buffer, s );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in ln_asbuf(): allocated buffer <%s>", l->buffer );
   bwb_debug( bwb_ebuf );
#endif

   /* strip CR from the buffer */

   bwb_stripcr( l->buffer );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in ln_asbuf(): stripped CRs" );
   bwb_debug( bwb_ebuf );
#endif

   return TRUE;

   }

/***************************************************************

        FUNCTION:       bwb_gets()

        DESCRIPTION:    This function reads a single line from
                        the specified buffer.

***************************************************************/

int
bwb_gets( char *buffer )
   {
   struct bwb_variable *v;
   char tbuf[ MAXSTRINGSIZE + 1 ];
   number = 0;

   v = var_find( DEFVNAME_PROMPT );
   str_btoc( tbuf, var_getsval( v ) );
   bwx_input( tbuf, buffer );

   return TRUE;
   }



/***************************************************************

        FUNCTION:       bwb_mainloop()

        DESCRIPTION:    This function is called 

	AGENDA:

***************************************************************/

void
bwb_mainloop( void )
   {
#if INTERACTIVE
   while (( exsc > -1 ) || !feof( stdin ))
      {
/*      bwb_execline(); */
      bwb_iteration();
      }
#else
   while ( exsc > -1 )
      {
/*      bwb_execline(); */
      bwb_iteration();
      }
#endif
   }

/***************************************************************

        FUNCTION:       break_mes()

        DESCRIPTION:    This function is called (a) by a SIGINT
			signal or (b) by error-handling routines.
			It prints an error message then calls
			break_handler() to handle the program
			interruption.

			JBV NOTE: No longer called by SIGINT
			signals or error-handling routines, but
			instead by bwb_iteration() since program
			interruption handling was moved there.
			break_handler() is also no longer called
			here since that has now been relegated
			to error handling to support CONT/RESUME.
			Finally, function prototype was changed
			from break_mes( int x ) to break_mes().
			(7/98)

***************************************************************/

void
break_mes( void )
   {
   static char *tmp_buffer;
   static int init = FALSE;

   /* get memory for temporary buffer if necessary */

   if ( init == FALSE )
      {
      init = TRUE;

      /* Revised to CALLOC0 pass-thru call by JBV */
      if ( ( tmp_buffer = (char *) CALLOC0( MAXSTRINGSIZE + 1, sizeof( char ), "break_mes")) == NULL )
         {
#if PROG_ERRORS
	 bwb_error( "in break_mes(): failed to find memory for tmp_buffer" );
#else
	 bwb_error( err_getmem );
#endif
	 }
      }

   /*--------------------------------------------*/
   /* Not yet, anyway; wait until bwb_iteration() */
   /* JBV 7/98                                   */
   /*--------------------------------------------*/
   /* expsc = 0;                         */
   /*--------------------------------------------*/

   /* Added leading \n (JBV 7/98) */
   sprintf( tmp_buffer, "\n\r%s %d\n", MES_BREAK, number );
   prn_xprintf( errfdevice, tmp_buffer );

   /*--------------------------------------------*/
   /* Removed by JBV 7/98 to support CONT/RESUME */
   /*--------------------------------------------*/
   /* break_handler();                           */
   /*--------------------------------------------*/

   }

/***************************************************************

        FUNCTION:       break_grabber()

        DESCRIPTION:    This function is called by a SIGINT signal.
			It sets the flag break_detected high.
			Later, after the current statement is done
			executing, the program interruption is
			handled in bwb_iteration().
			Added by JBV 7/98 to support CONT/RESUME.

***************************************************************/

void
break_grabber( int x )
   {

#if INTERACTIVE

   /* reset the break handler */

#if HAVE_SIGNAL
   signal( SIGINT, break_grabber );
#endif

   break_detected = 1;

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in break_grabber(): Break detected" );
   bwb_debug( bwb_ebuf );
#endif

#else		/* nonINTERACTIVE:  terminate immediately */

   bwx_terminate();

#endif

   }

/***************************************************************

	FUNCTION:       break_handler()

	DESCRIPTION:    This function is called by break_mes()
			and handles program interruption by break
			(or by the STOP command).

			JBV NOTE: No longer used by break_mes()
			or for general program interruptions.
			Since CONT/RESUME functionality was added,
			it is only used now for program interruptions
			related to errors. (7/98)

***************************************************************/

void
break_handler( void )
   {

#if INTERACTIVE		/* INTERACTIVE: reset counters and jump back to mark */

   /* reset all stack counters */

   exsc = -1;
   expsc = 0;
   xtxtsc = 0;
   err_gosubl[ 0 ] = '\0';

   /* reset the break handler */

#if HAVE_SIGNAL
   /* Interrupt handler was break_mes (JBV 7/98) */
   signal( SIGINT, break_grabber );
#endif

#if HAVE_LONGJUMP
   longjmp( mark, -1 );
#else		/* HAVE_LONGJUMP = FALSE; no jump available; terminate */
   bwx_terminate();
#endif

#else		/* nonINTERACTIVE:  terminate immediately */

   bwx_terminate();

#endif

   }


/***************************************************************

	FUNCTION:       is_ln()

	DESCRIPTION:    This function determines whether a program
			line (in buffer) begins with a line number.

***************************************************************/

int
is_ln( char *buffer )
   {
   static int position;

   position = 0;
   adv_ws( buffer, &position );
   switch( buffer[ position ] )
      {
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
         return TRUE;
      default:
         return FALSE;
      }
   }


/***************************************************************

	FUNCTION:       CALLOC0()

	DESCRIPTION:    Pass-thru function to calloc() for debugging
			purposes.  Added by JBV 10/95

***************************************************************/

void *
CALLOC0( size_t nelem, size_t elsize, char *str )
{
    void *ptr;

    ptr = calloc(nelem, elsize);
    /* printf("%x %x\n", ptr, mallocblksize(ptr)); */
    return ptr;
}

/***************************************************************

	FUNCTION:       FREE0()

	DESCRIPTION:    Pass-thru function to free() for debugging
			purposes.  Added by JBV 10/95

***************************************************************/

void
FREE0( void *ptr, char *str )
{
    /* printf("%x\n", ptr); */
    free(ptr);
}


