/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 * File: main.c
 *
 * Created: Wed Dec 11 13:31:43 2002
 *
 * $Id: main.c,v 1.1 2003/12/13 20:59:15 mhassler Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "defines.h"
#include "session.h"
#include "dns.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif /* DMALLOC */

#define SETSIG(sa, sig, fun)	sa.sa_handler = fun; \
				sa.sa_flags = SA_RESTART; \
				sigaction(sig, &sa, 0L);

int quit = 0;

extern struct session *session_list;

int filter_proxy = 0;
int filter_cookies = 0;
int filter_java = 0;
int filter_activex = 0;
int filter_popups = 0;

void terminate(int p)
{
	dprintf("Received signal %i, exiting...\n", p);
	quit = 1;
}

void usage()
{
  dprintf("Usage: foxy OPTION\n\n");
	dprintf("Options:\n\t-h\tThis help screen\n");
	dprintf("\t-y\tFilter HTTP proxy\n");
	dprintf("\t-c\tFilter cookies\n");
	dprintf("\t-j\tFilter Java applets\n");
	dprintf("\t-a\tFilter ActiveX\n");
	dprintf("\t-p\tFilter Popup windows\n");
}

void parameters(int argc, char *argv[])
{
	int c;
	char help = 0, par_error = 0;

	while(1)
	{
		c = getopt(argc, argv, "hycjap");
		if(c == -1)
			break;
		switch(c)
		{
			case 'h':
				help = 1;
				break;
			case 'y':
				dprintf("Filtering HTTP proxy\n");
				filter_proxy = 1;
				break;
			case 'c':
				dprintf("Filtering cookies\n");
				filter_cookies = 1;
				break;
			case 'j':
				dprintf("Filtering Java applets\n");
				filter_java = 1;
				break;
			case 'a':
				dprintf("Filtering ActiveX\n");
				filter_activex = 1;
				break;
			case 'p':
				dprintf("Filtering Popup windows\n");
				filter_popups = 1;
				break;
			default:
				par_error = 1;
				break;
		}
	}
	if(help || par_error)
	{
		usage();
		exit(0);
	}
/*	if(!filter_proxy && !filter_cookies && !filter_java && !filter_activex && !filter_popups)
	{
		usage();
		exit(0);
	}*/
}

int foxy_listen()
{
	struct sockaddr_in sin;
	int sok, opt;

	sok = socket(AF_INET, SOCK_STREAM, 0);
	if(sok < 0)
		return -1;

	if(setsockopt(sok, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0)
		return -1;

	sin.sin_family = AF_INET;
	sin.sin_port = htons(FOXY_PORT);
	sin.sin_addr.s_addr = 0; //htonl(FOXY_ADDR);
	if(bind(sok, (struct sockaddr*)&sin, sizeof(struct sockaddr_in)) < 0)
		return -1;
	if(listen(sok, 5) < 0)
		return -1;
	return sok;
}

int main(int argc, char **argv)
{
	struct sigaction sa;
	int res, maxfd = 0, lstn_sok;
	fd_set rfds, wfds;
	struct timeval tm;
	struct session *ptr;

#if 0
	res = dns_query("www.yahoo.com", 13);
	dns_read(res);
	dprintf("=================\n");
	res = dns_query("www.slashdot.org", 16);
	dns_read(res);
	exit(0);
#endif

	parameters(argc, argv);

	lstn_sok = foxy_listen();
	if(lstn_sok < 0)
	{
		perror("bind");
		return 1;
	}

	SETSIG(sa, SIGTERM, terminate);
	SETSIG(sa, SIGINT, terminate);

//	close(0);
//	close(1);
//	close(2);

	while(!quit)
	{
		tm.tv_sec = 5;
		tm.tv_usec = 0;

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);

		// Only care about new incoming connections if not maxed out
		if(session_count() < MAX_SESSIONS)
		{
			FD_SET(lstn_sok, &rfds);
			maxfd = lstn_sok;
		}

		dprintf("%d sessions: ", session_count());
		ptr = session_list;
		while(ptr)
		{
			dprintf("%d,", ptr->id);
			ptr = ptr->next;
		}
		dprintf("\n");

		session_fill_rfds(&rfds, &maxfd);
		session_fill_wfds(&wfds, &maxfd);

		res = select(maxfd+1, &rfds, &wfds, 0, &tm);
		if(res < 0)
		{
			// error or signal
		}
		else if(res == 0)
		{
			// timeout handling
		}
		else
		{
			// something on a socket
			if(FD_ISSET(lstn_sok, &rfds))
			{
				dprintf("New connection!\n");
				session_accept(lstn_sok);
			}
			session_iterate_read(&rfds);
			session_iterate_write(&wfds);
		}
		session_iterate_parse_client();
		session_iterate_parse_server();
		session_iterate_delete();
	}

	session_iterate_shutdown();
	close(lstn_sok);

	return 0;
}
