/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 * File: session.h
 *
 * Created: Wed May 14 15:37:26 2003
 *
 * $Id: session.h,v 1.1 2003/12/13 20:59:15 mhassler Exp $
 */

#ifndef _SESSION_H_
#define _SESSION_H_

#include <sys/select.h>

#define ST_METHOD 0 // GET/POST, etc
#define ST_STATUS 0 // 200 OK, etc
#define ST_HEADER 1 // Host:, Content-Type:, etc
#define ST_BODY   2

#define MET_ERROR     -1
#define MET_UNKNOWN   0
#define MET_OPTIONS   1
#define MET_GET       2
#define MET_HEAD      3
#define MET_POST      4
#define MET_PUT       5
#define MET_DELETE    6
#define MET_TRACE     7
#define MET_CONNECT   8

#define BST_HTML    0
#define BST_BINARY  1
#define BST_APPLET  2
#define BST_OBJECT  3
#define BST_SCRIPT  4
#define BST_BODYTAG 5

struct buffer
{
	char *data;
	int size;
};

struct stack
{
	int state;

	struct stack *next;
	struct stack *prev;
};

struct connection
{
	int sok;
	int method;
	int state;
	struct stack *bodystate;
	int skip;
	int parsable;
	long cont_len;
	int chunked;
	int chunksize;
	unsigned long ip;    // HOST order
	unsigned short port; // HOST order

	struct buffer in;
	struct buffer out;
};

struct session
{
	struct connection client;
	struct connection server;

	int dns_sok;
	int id;

	struct session *next;
	struct session *prev;
};

void push_state(struct connection *con, int state);
int pop_state(struct connection *con);
int peek_state(struct connection *con);

int session_count();
int session_accept(int lsok);

int session_close(struct session *ses);
int session_close_server(struct session *ses);

int session_fill_rfds(fd_set *rfds, int *maxfd);
int session_fill_wfds(fd_set *wfds, int *maxfd);

int session_iterate_read(fd_set *rfds);
int session_iterate_write(fd_set *fds);

int session_iterate_parse_client();
int session_iterate_parse_server();

int session_iterate_delete();

int session_read_client(struct session *ses);
int session_write_connection(struct connection *conn);
int session_read_server(struct session *ses);

int client_parse(struct session *ses);
int server_parse(struct session *ses);

int session_iterate_shutdown();

#endif
