#!/bin/bash
#---------------------------------------------------------------
# This script is called from both the ROHS and standard firmware makefiles
# "FILESYS_DIR" should be defined in them as appropriate
#---------------------------------------------------------------

# ------------------------------------
# change this function to do whatever you need to do to compile the sources
compile_it()
{
   make -i clean
   ./configure --host=mips --build==mipsel-unknown-linux-gnu --with-config=/etc/tinyproxy.conf --enable-xtinyproxy --enable-filter --enable-upstream --enable-reverse --enable-transparent-proxy
   make -i
}

# ------------------------------------
# where to copy things - default, if none is supplied in the calling Makefile
if test "$FILESYS_DIR" = ""; then
   FILESYS_DIR=../../filesystem
fi

# program to build
if test "$BUILD_PROGR" = ""; then
   BUILD_PROGR=src/tinyproxy
fi

# where to install it to
if test "$DEST_DIREC" = ""; then
   DEST_DIREC=$FILESYS_DIR/usr/local/bin
fi

# do we want to install too?
if test "$1" = "install" || test "$2" = "install" || test "$3" = "install" || test "$4" = "install"; then
   TO_INSTALL=1
else
   TO_INSTALL=0
fi

# is this an ROHS build? (currently unused)
if test "$1" = "rohs" || test "$2" = "rohs" || test "$3" = "rohs" || test "$4" = "rohs"; then
   ROHS=1
else
   ROHS=0
fi

# which OpenWrt gcc toolchain version do we want?
if test "$GCC_BASE" = "3.4.6"; then
   gcc_ver=3.4.6
else
   gcc_ver=4.1.2
fi

# --- fix the paths
CROSS_COMPILE=mipsel-uclibc
CFLAGS="-Os -s -ffunction-sections -fdata-sections"
LDFLAGS="-Wl,--gc-sections"
AS=$CROSS_COMPILE-as
LD=$CROSS_COMPILE-ld
RANLIB=$CROSS_COMPILE-ranlib
CPP="$CROSS_COMPILE-gcc -E"
AR=$CROSS_COMPILE-ar
NM=$CROSS_COMPILE-nm
STRIP=$CROSS_COMPILE-strip
OBJCOPY=$CROSS_COMPILE-objcopy
OBJDUMP=$CROSS_COMPILE-objdump
HOSTCC=/usr/bin/gcc

BASEDIR=/opt/gcc/ar7/staging_dir
export PATH=$BASEDIR/toolchain-mipsel_gcc$gcc_ver/bin:$BASEDIR/host/bin:$PATH
export AS LD CC CPP AR NM STRIP OBJCOPY OBJDUMP CFLAGS RANLIB LDFLAGS HOSTCC

# cleanup old builds ?
if test "$1" = "clean" || test "$2" = "clean" || test "$3" = "clean" || test "$4" = "clean"; then
   make distclean
   cd src
   make clean 
   rm -f tinyproxy
   cd ..
   exit
fi

# if not install, prepare to build
if test "$TO_INSTALL" = "0"; then
   compile_it;
fi

# install
if test "$TO_INSTALL" = "1"; then
   if ! test -e "$BUILD_PROGR"; then
      compile_it;
   fi
   install -d $DEST_DIREC
   install -d $FILESYS_DIR/etc
   install -d $FILESYS_DIR/usr/www/tinyproxy
   install -m 0755 src/tinyproxy-bin $DEST_DIREC/tinyproxy
   install -m 0755 src/tinyproxy $DEST_DIREC/tinyproxy-bin
   install -m 0755 src/tinytrans.sh $DEST_DIREC/tinytrans.sh
   ln -sf ../var/tmp/tinyproxy.filter $FILESYS_DIR/etc/tinyproxy.filter
   cp -f -a doc/*.html $FILESYS_DIR/usr/www/tinyproxy/
fi

